/*
 *   setmix - Audio Mixer Setup Program
 *
 *   Author: Gaspar Sinai (gsinai@gol.com) 
 *   Copyright (C) 1996  Gaspar Sinai
 * 
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   This program is setting the mixer values from a dump file created
 *   by either this program or xmmix. It is useful if you want to save 
 *   the setting for users logging in and out at the console. They can 
 *   receive  their own defults of sound setting.
 * 
 *   Many thanks to Ti Kan (ti@amb.org) who made xmmix available. Most
 *   of the code in this program is based upon that excellent application.
 *
 *   xmmix is a GNU Copyrighted Software of Ti Kan.
 *
 */


/* I like to put all definitions in the header files */
#define __setmix_c_

#define VERSION "0.6"

/*
 *  To make users  of the "what" program happy.
 */
static char *what_id = "@(#)setmix 0.6  31 January 1998.";

#include "setmix.h"

#if defined(SOUND_MIXER_LOUD) && defined(SOUND_MIXER_MUTE) && \
    defined(SOUND_MIXER_ENHANCE)
#if SOUND_MIXER_LOUD == SOUND_MIXER_MUTE
/* HACK for new sound drivers */
#undef SOUND_MIXER_LOUD
#undef SOUND_MIXER_MUTE
#undef SOUND_MIXER_ENHANCE
#endif
#endif


char 	*Me="?";	/* The basename of this program */
bool_t	Debug=0;	/* Debug level */
static  dev_fd;		/* Mixer Device File descriptor */
int	maxdevs=0;      /* Max Number of devices */
char	drv_ver[128];	/* Sound driver version */

/*
 * Well, I wnated to name this funtion to something else but all names
 * were already in use.
 */
main (int NumArgs, char **Args)
{
	int		i;
	int		option;
	int		write = -1;
	char		filename[FILENAME_MAX+1];
	char		*home;
	char		*device = "/dev/mixer";

	/* to be or not to be */
	Me=Args[0];
	filename[0]=0;

	/* get the basename */
	for (i=0; Args[0][i]; i++) if (Args[0][i] == '/') Me=&Args[0][i+1];
	
	/* If we block some guys during boot will curse us */ 
	signal (SIGALRM, my_sighandler);
	alarm (20);

	home = getenv ("HOME");
	if (home==NULL) home=".";

	/* I was first I got the name */
	strcpy (filename, home);
	strcat (filename, "/.mixrc");

	write =  (!strcmp(Me, "getmix")) ? 1 : (!strcmp(Me, "setmix")) ? 0 : -1;

	/* Option Processing w=write l=load f<filename> */
	while ((option=getopt (NumArgs, Args, "gsf:d:x")) > 0) switch (option)
	{
		case 'g': write=1; break;
		case 's': write=0; break;
		case 'f': strcpy(filename,optarg?optarg:"");break;
		case 'd': device=optarg;break;
		case 'x': Debug=TRUE;break;
		default: exit(1);
	}

	if ( write < 0)
	{
		fprintf (stderr, "usage: setmix [-g|-s] [-f filepathname] [-d device] [-x]\n");
		exit (255);
	}
	if (optind!=NumArgs)
	{
		fprintf (stderr, "usage: %s [-g|-s] [-f filepathname] [-d device] [-x]\n", Me);
		exit (255);
	}

	if (!init_mixer (device))
	{
		deinit_mixer (device);
		exit(255);
	}

	/* Switch is beautiful */
	switch (write) 
	{
	case 0:
		if (!load_mixer(filename)) 
		{
			fprintf(stderr, "Cannot read mixer settings file:\n%s\n",
				filename);
			deinit_mixer (device);
			exit (255);
		}
		break;
	default:
		if (!write_mixer(filename)) 
		{
			fprintf(stderr, "Cannot write mixer settings file:\n%s\n",
				filename);
			deinit_mixer (device);
			exit (255);
		}
		break;
	}
	deinit_mixer ();
	exit (0);
}

/* 
 * init_mixer 
 * Do:     Query sound board mixer feature capabilities and current hardware 
 *         settings.  
 * Return: Nonzero on success
 */ 
static bool_t 
init_mixer (char *device) 
{ 
	int	i; 
	int	level; 
	int	curmask;
	char	errmsg[STR_BUF_SZ];
	char 	iocstr[STR_BUF_SZ];
	FILE	*fp;
	char	*p;
	char	tmpbuf[128];

	/* Open device */
	sprintf(errmsg, "%s: Cannot open device %s",
		Me, device);

	if ((dev_fd = open(device, O_RDONLY)) < 0) 
	{
		perror(errmsg);
		return FALSE;
	}

	/* Read device mask */
	do_ioctl
	(
		SOUND_MIXER_READ_DEVMASK,
		&ctlinfo.devmask,
		"SOUND_MIXER_READ_DEVMASK",
		IOC_DATA_IN
	);

	/* Read record mask */
	do_ioctl(
		SOUND_MIXER_READ_RECMASK,
		&ctlinfo.recmask,
		"SOUND_MIXER_READ_RECMASK",
		IOC_DATA_IN
	);

	/* Read stereo devices */
	do_ioctl(
		SOUND_MIXER_READ_STEREODEVS,
		&ctlinfo.stereodevs,
		"SOUND_MIXER_READ_STEREODEVS",
		IOC_DATA_IN
	);

#ifdef SOUND_MIXER_CAPS
	/* Check mixer capability */
	do_ioctl(
		SOUND_MIXER_READ_CAPS,
		&ctlinfo.caps,
		"SOUND_MIXER_READ_CAPS",
		IOC_DATA_IN
	);
#endif

	/*
	 * This part will have to be modified if a new version of sound
	 * driver appears. 
         */

	/* Set maximum number of devices supported */
	maxdevs = MAXDEVS_V3;

	strcpy(drv_ver, "unknown");

	/* Query sound driver version */
	if ((fp = fopen("/dev/sndstat", "r")) == NULL) 
	{
		if (Debug) 
		{
			fprintf(stderr, "%s: %s: %s\n",
				Me,
				"Cannot open /dev/sndstat",
				"assuming pre-3.0 VoxWare driver.");
		}
		maxdevs = MAXDEVS_V2;
		/* Yes !*/
		return TRUE;
	}

	if (fgets(tmpbuf, sizeof(tmpbuf), fp) == NULL) 
	{
		if (Debug) 
		{
			fprintf(stderr, "%s: %s: %s\n",
				Me,
				"Cannot read /dev/sndstat",
				"assuming pre-3.0 VoxWare driver.");
		}
		maxdevs = MAXDEVS_V2;
		fclose(fp);
		/* Yes !*/
		return TRUE;
	}

	fclose(fp);

	/* Eat newline */
	tmpbuf[strlen(tmpbuf) - 1] = '\0';

	/* Skip forward to the first colon character */
	for (p = tmpbuf; *p != '\0' && *p != ':'; p++)
		;

	if (*p != ':') 
	{
		if (Debug) 
		{
			fprintf(stderr, "%s: %s: %s\n",
				Me,
				"Cannot parse info from /dev/sndstat",
				"assuming pre-3.0 VoxWare driver.");
		}
		maxdevs = MAXDEVS_V2;
		strcpy(drv_ver, tmpbuf);
		return TRUE;
	}

	/* Skip blanks */
	p++;
	while (*p == ' ' || *p == '\t')
		p++;

	strcpy(drv_ver, p);

	if (*p < '3')
		/* Running on VoxWare 2.x: set max devices accordingly */
		maxdevs = MAXDEVS_V2;

	if (Debug) 
	{
		fprintf(stderr, "Sound driver [%s]\n", drv_ver);
		fprintf(stderr, "Running with maxdevs=%d\n\n", maxdevs);
	}
	return TRUE;
}

/* 
 * deinit_mixer 
 * Do:     Deinitialize device
 * Return: Nonzero on success
 */ 
static bool_t 
deinit_mixer () 
{ 
	/* Close device */
	return (bool_t) !close(dev_fd);
}

/*
 * load_mixer
 * Do:     Read the mixer settting of a given file and update settings
 * Return: Nonzero on success
 */
static bool_t
load_mixer(char *path)
{
	FILE	*fp;
	int	i,
		curmask,
		val1,
		val2,
		val3,
		val4;
	char	*p,
		buf[STR_BUF_SZ];

	/* Open file for reading */
	if ((fp = fopen(path, "r")) == NULL)
		return FALSE;

	/* Read first line of mixer settings file */
	if (fgets(buf, sizeof(buf), fp) == NULL) 
	{
		fclose(fp);
		return FALSE;
	}

	/* Mixer settings file signature check */
	if (strncmp(buf, "# xmmix ", 8) != 0 ) 
	{
		fclose(fp);
		return FALSE;
	}

	/* Read the rest of the mixer settings file */
	while (fgets(buf, sizeof(buf), fp) != NULL) 
	{
		if (buf[0] == '#')
			/* Comment line */
			continue;

		if ((p = strchr(buf, '=')) == NULL)
			/* Invalid line */
			continue;

		*p = '\0';

		if (sscanf(p+1, "%d,%d,%d,%d\n",
			   &val1, &val2, &val3, &val4) == 4) 
		{

			for (i = 0; i < maxdevs; i++) 
			{
				if (strcmp(buf, ctlinfo.slinfo[i].name) == 0)
					break;
			}

			if (i >= maxdevs)
				/* Unrecognized keyword */
				continue;

			curmask = (1 << i);

			if (!(ctlinfo.devmask & curmask))
				/* Device not supported */
				continue;

			ctlinfo.slinfo[i].left = val1;
			ctlinfo.slinfo[i].right = val2;

			if (val1 == val2)
				ctlinfo.slinfo[i].locked = (bool_t) (val3 > 0);
			else
				ctlinfo.slinfo[i].locked = FALSE;

			if (val4 > 0 && (ctlinfo.recmask & curmask))
				ctlinfo.recsrc |= (1 << i);
			else
				ctlinfo.recsrc &= ~(1 << i);
		}

#ifdef SOUND_ONOFF_MAX

		else if (sscanf(p+1, "%d\n", &val1) == 1) 
		{
#ifdef SOUND_MIXER_MUTE
			if (strcmp(buf, "mute") == 0 &&
			    ctlinfo.devmask & (1 << SOUND_MIXER_MUTE)) 
			{
				ctlinfo.mute = (bool_t) (val1 > 0);
			}
#endif /* SOUND_MIXER_MUTE */
#ifdef SOUND_MIXER_ENHANCE
			if (strcmp(buf, "enhance") == 0 &&
				 ctlinfo.devmask & (1 << SOUND_MIXER_ENHANCE)) 
			{
				ctlinfo.enhance = val1;
			}
#endif /* SOUND_MIXER_ENHANCE */
#ifdef SOUND_MIXER_LOUD
			if (strcmp(buf, "loudness") == 0 &&
				 ctlinfo.devmask & (1 << SOUND_MIXER_LOUD)) 
			{
				ctlinfo.loudness = (bool_t) (val1 > 0);
			}
#endif /* SOUND_MIXER_LOUD */
		}
#endif /* SOUND_ONOFF_MAX */
	}

	fclose(fp);

	/* Update controls to match current status */
	return set_device();
}


/*
 * write_mixer
 * Do:     Write the mixer settting into a given file
 * Return: Nonzero on success
 */
static bool_t
write_mixer(char *path)
{
	FILE	*fp;
	int	i;
	int	count;

	/* Open file for writing */
	if ((fp = fopen(path, "w")) == NULL)
	{
		return FALSE;
	}

        /* get the current settings into ctlinfo */
	if( !get_device())
	{
		fprintf (stderr, "%s: Can not get device settings.\n");
		return FALSE;
	}

	/* Write first two lines of mixer settings file */
	fprintf(fp, "# xmmix (setmix %s) Mixer Settings File.\n", 
			VERSION);
	fprintf(fp, "# Compiled with Unix Sound System version %d.\n", 
			SOUND_VERSION);
	fprintf(fp, "# Syntax:\n" );
	fprintf(fp, "#         name=left,right,locked,recsource\n" );
	fprintf(fp, "# Currently the following names are recognized:" );

	for (i=0, count=0; i<MAXDEVS;i++)
	{
		if (ctlinfo.slinfo[i].name == NULL) continue;
		if (i%8 == 0) fprintf (fp, "\n#         ");
		fprintf (fp, "%s ", ctlinfo.slinfo[i].name);
	}
	fprintf (fp, "\n");

	/* Write all control settings */
	for (i = 0; i < MAXDEVS; i++) 
	{
		if ( ctlinfo.slinfo[i].name == NULL)
			break;

		if (ctlinfo.slinfo[i].left < 0 || ctlinfo.slinfo[i].right < 0)
			continue;
		fprintf(fp, "%s=%d,%d,%d,%d\n",
			ctlinfo.slinfo[i].name,
			ctlinfo.slinfo[i].left,
			ctlinfo.slinfo[i].right,
			ctlinfo.slinfo[i].locked,
			(int) ((ctlinfo.recsrc & (1 << i)) > 0));
	}

#ifdef SOUND_ONOFF_MAX
	if (ctlinfo.mute >= 0)
		fprintf(fp, "%s=%d\n", "mute", (int) ctlinfo.mute);

	if (ctlinfo.enhance >= 0)
		fprintf(fp, "%s=%d\n", "enhance", ctlinfo.enhance);

	if (ctlinfo.loudness >= 0)
		fprintf(fp, "%s=%d\n", "loudness", (int) ctlinfo.loudness);
#endif

	return (fclose(fp) == 0);
}


/*
 * get_device
 * Do:     Query mixer settings and update current ctlinfo state.
 * Return: FALSE on error
 */
static int
get_device()
{
	int		i;
	int		level;
	int		curmask;
	char		iocstr[STR_BUF_SZ];
	int		status;
		
	/* Read record source */
	do_ioctl(
		SOUND_MIXER_READ_RECSRC,
		&ctlinfo.recsrc,
		"SOUND_MIXER_READ_RECSRC",
		IOC_DATA_IN
	);

	/* Read control settings */
	for (i = 0; i < maxdevs; i++) 
	{
		curmask = (1 << i);

		sprintf(iocstr, "MIXER_READ[%d]:%s", i,  ctlinfo.slinfo[i].name);
		if( !do_ioctl(MIXER_READ(i), &level, iocstr, IOC_DATA_IN)) 
		{
			ctlinfo.slinfo[i].left = ctlinfo.slinfo[i].right = -1;
			continue;
		}

		ctlinfo.slinfo[i].left = (level & 0xff);
		ctlinfo.slinfo[i].right = ((level >> 8) & 0xff);

		/* Sanity check */
		if (ctlinfo.slinfo[i].left > 100)
			ctlinfo.slinfo[i].left = 100;
		else if (ctlinfo.slinfo[i].left < 0)
			ctlinfo.slinfo[i].left = 0;
		if (ctlinfo.slinfo[i].right > 100)
			ctlinfo.slinfo[i].right = 100;
		else if (ctlinfo.slinfo[i].right < 0)
			ctlinfo.slinfo[i].right = 0;

		if ((curmask & ctlinfo.stereodevs) == 0) 
		{
			ctlinfo.slinfo[i].left =
				ctlinfo.slinfo[i].right =
				(ctlinfo.slinfo[i].left +
				 ctlinfo.slinfo[i].right) / 2;
			ctlinfo.slinfo[i].locked = TRUE;
		}

		if (ctlinfo.slinfo[i].left != ctlinfo.slinfo[i].right)
			ctlinfo.slinfo[i].locked = FALSE;
	}

#ifdef SOUND_ONOFF_MAX
	/* Additional toggle capabilitites */
	for (i = SOUND_ONOFF_MIN; i <= SOUND_ONOFF_MAX; i++) 
	{
		curmask = (1 << i);

		if (curmask & ctlinfo.devmask) 
		{
			sprintf(iocstr, "MIXER_READ[%d]:%s", i, ctlinfo.slinfo[i].name);
			status=do_ioctl(MIXER_READ(i), &level, iocstr, IOC_DATA_IN);

			switch (i) {
#ifdef SOUND_MIXER_MUTE
			case SOUND_MIXER_MUTE:
				if(status == TRUE) 
					ctlinfo.mute = (level != 0);
				else
					ctlinfo.mute = -1;
				break;
#endif /* SOUND_MIXER_MUTE */
#ifdef SOUND_MIXER_LOUD
			case SOUND_MIXER_LOUD:
				if(status == TRUE) 
					ctlinfo.loudness = (level != 0);
				else
					ctlinfo.loudness = -1;
				break;
#endif /* SOUND_MIXER_LOUD */
#ifdef SOUND_MIXER_ENHANCE
			case SOUND_MIXER_ENHANCE:
				/* Hack: The enhance values are hard wired. */
				if(status == FALSE) 
				{
					ctlinfo.enhance = -1;
					break;
				}
				switch (level) {
				case 80:
					ctlinfo.enhance = 3;
					break;
				case 60:
					ctlinfo.enhance = 2;
					break;
				case 40:
					ctlinfo.enhance = 1;
					break;
				case 0:
				default:
					ctlinfo.enhance = 0;
					break;
				}
				break;
#endif /* SOUND_MIXER_ENHANCE */
			default:
				/* Unsupported feature: ignore */
				break;
			}
		}
	}
#endif	/* SOUND_ONOFF_MAX */

	/* No other return value yet */
	return TRUE;
}

/*
 * set_device
 * Do:     Set the device to whatever is in ctlinfo structure
 * Return: FALSE on error
 */
static int
set_device()
{
	int		i;
	int		curmask;
	int		level;
	char		iocstr[STR_BUF_SZ];

	/* Read record source */
	do_ioctl(
		SOUND_MIXER_WRITE_RECSRC,
		&ctlinfo.recsrc,
		"SOUND_MIXER_WRITE_RECSRC",
		IOC_DATA_IN
	);

	/* Read control settings */
	for (i = 0; i < maxdevs; i++) 
	{
		curmask = (1 << i);
		curmask = (1 << i);

		if (ctlinfo.slinfo[i].left < 0 || ctlinfo.slinfo[i].right < 0)
			continue;
			
		/* Sanity check */
		if (ctlinfo.slinfo[i].left > 100)
			ctlinfo.slinfo[i].left = 100;
		if (ctlinfo.slinfo[i].right > 100)
			ctlinfo.slinfo[i].right = 100;

		if ((curmask & ctlinfo.stereodevs) == 0) 
		{
			ctlinfo.slinfo[i].left =
				ctlinfo.slinfo[i].right =
				(ctlinfo.slinfo[i].left +
				 ctlinfo.slinfo[i].right) / 2;
			ctlinfo.slinfo[i].locked = TRUE;
		}

		level=ctlinfo.slinfo[i].left + (ctlinfo.slinfo[i].right << 8);
		sprintf(iocstr, "MIXER_WRITE[%d]:%s", i, ctlinfo.slinfo[i].name);
		do_ioctl(MIXER_WRITE(i), &level, iocstr, IOC_DATA_OUT);
	}

#ifdef SOUND_ONOFF_MAX
	/* Additional toggle capabilitites */
	for (i = SOUND_ONOFF_MIN; i <= SOUND_ONOFF_MAX; i++) 
	{
		curmask = (1 << i);

		if (curmask & ctlinfo.devmask) 
		{

			sprintf(iocstr, "MIXER_WRITE[%d]:%s", 
				i, ctlinfo.slinfo[i].name);
			switch (i) {
#ifdef SOUND_MIXER_MUTE
			case SOUND_MIXER_MUTE:
				if (ctlinfo.mute<0) break;
				do_ioctl(MIXER_WRITE(i), &ctlinfo.mute, iocstr, 
					IOC_DATA_OUT);
				break;
#endif /* SOUND_MIXER_MUTE */
#ifdef SOUND_MIXER_LOUD
			case SOUND_MIXER_LOUD:
				if (ctlinfo.loudness<0) break;
				do_ioctl(MIXER_WRITE(i), &ctlinfo.loudness, iocstr, 
					IOC_DATA_OUT);
				break;
#endif /* SOUND_MIXER_LOUD */
#ifdef SOUND_MIXER_ENHANCE
			case SOUND_MIXER_ENHANCE:
				/* Hack: The enhance values are hard wired. */
				if (ctlinfo.enhance < 0) break;
				switch (ctlinfo.enhance) {
				case 3:
					level = 80;
					do_ioctl(MIXER_WRITE(i),&level, 
						iocstr, IOC_DATA_OUT);
					break;
				case 2:
					level = 60;
					do_ioctl(MIXER_WRITE(i),&level, 
						iocstr, IOC_DATA_OUT);
					break;
				case 1:
					level = 40;
					do_ioctl(MIXER_WRITE(i),&level, 
						iocstr, IOC_DATA_OUT);
					break;
				case 0:
				default:
					ctlinfo.enhance = 0;
					break;
				}
				break;
#endif /* SOUND_MIXER_ENHANCE */
			default:
				/* Unsupported feature: ignore */
				break;
			}
		}
	}
#endif	/* SOUND_ONOFF_MAX */

	/* No other return value yet */
	return TRUE;
}

/*
 * do_ioctl
 * Do:      Perform ioctl command. If file is not yet open just quietly return.    
 * Return:  False on failure.
 * Args:
 *	cmd - The ioctl command
 *	arg - The ioctl argument
 *	name - The ioctl command string
 *	dir - Data direction
 *		IOC_DATA_IN	
 *		IOC_DATA_OUT
 *		IOC_DATA_NONE
 */
static int
do_ioctl(int cmd, int *arg, char *name, int dir)
{
	int	ret;
	char	errmsg[STR_BUF_SZ];

	if (dev_fd < 0)
	{
		return FALSE;
	}

	sprintf(errmsg, "\n%s ioctl failed", name);

	if ((ret = ioctl(dev_fd, cmd, arg)) < 0)
	{
		/* Some may fail to exist */
		if (Debug) perror(errmsg);
		return FALSE;
	}
	return TRUE;
}

/*
 * My sighandler is envoked if some terrible thing has happened.
 */
void
my_sighandler(int signal)
{
	char 	*sigtype; 
	switch(signal)
	{
	case SIGALRM: sigtype="SIGALRM";break;
	default:
		fprintf (stderr, "%s: goind down on signal %d.\n", Me, signal);
		exit(1);
	}
	fprintf (stderr, "%s: goind down on signal %s.\n", Me, sigtype);
	exit(1);
}
